Attribute VB_Name = "MainProcs"
Option Explicit
Public Declare Function BitBlt Lib "gdi32" (ByVal hDestDC As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal dwRop As Long) As Long

Public Const cBITBLT_PATCOPY As Long = &HF00021
Public Const cEDITGRID As Integer = 500
Public Const cRADIUS As Integer = 160
Public Const cSTART As String = "St"
Public Const cFINISH As String = "Fn"
Public Const cSELRADIUS As Integer = cRADIUS * cRADIUS

Private Const cMININT As Integer = -32767

Public gDirtyFile As Boolean
Public gLinks As Dictionary
Public gNodes As Dictionary
Public gNextKey As Integer

Private mForm As frmCriticalAnalyser
Private mFileName As String

Public Sub SetFileName(Action As eFileAction)
On Local Error GoTo CancelError
Dim SplitPath As Variant

    'Use a common dialog to get the XML path and filename. Use this to load up a DOM Document
    With mForm.Dlg
        .InitDir = App.Path
        .FileName = "*.xml"
        .DialogTitle = Choose(Action, "Open File", "Save File")
        Do
            Select Case Action
                Case eOpen: .ShowOpen
                Case eSaveAs: .ShowSave
            End Select
        Loop Until Not Len(.FileName) = 0
        'Get the name of the file for the form's caption
        SplitPath = Split(.FileName, "\")
        mForm.Caption = "Critical Path Analyser - " & SplitPath(UBound(SplitPath))
        'Set the current xml document path and name
        mFileName = .FileName
    End With

Exit Sub
CancelError:
    If Not Err = cdlCancel Then Err.Raise Err.Number, , Err.Description
End Sub

Public Sub FindCriticalPath()
Dim Index As Integer, Best As Integer, Dist As Integer
Dim Candidates As New Collection
Dim Nd As Node, ToNode As Node
Dim Links As Variant
Dim Lk As Link
    
    'This proc works out the longest path from the dummy start node to the dummy finsih node
    
    'Reset the objects ready to re-calculate the path
    CPClear
    'Prime the loop by making dummy start node the first candidate to be considered
    Candidates.Add gNodes(cSTART)
    'Keep looping until there are no more candidates to check
    Do While Candidates.Count > 0
        'Take the first candidate from the list, its as good as any
        Set Nd = Candidates(1)
        Best = Nd.Distance
        'Remove the candidate from the list
        Candidates.Remove 1
        'Get the links that radiate out from the candidate node
        Links = Nd.Links.Items
        'Loop around these links checking the node at the other end to see how far away it is
        For Index = 0 To UBound(Links)
            Set Lk = Links(Index)
            Set ToNode = Lk.Nd2
            Dist = Best + Lk.Length
            'If the node at the end of the link is farther away than the best so far then add it to the candidate list
            If Dist >= ToNode.Distance Then
                Candidates.Add ToNode
                ToNode.Distance = Dist
                'Set the InLink property to update the current 'Longest Path'
                Set ToNode.InLink = Lk
            End If
        Next
    Loop
    'Mark the objects to make them easy to draw
    CPMarkUp

End Sub

Public Sub CPMarkUp()
Dim Links As Variant, Nodes As Variant
Dim Index As Integer
Dim Nd As Node

    'Get the nodes and mark those that are in the tree so they can be colored blue.
    Nodes = gNodes.Items
    For Index = 0 To UBound(Nodes)
        If Not Nodes(Index).InLink Is Nothing Then Nodes(Index).InLink.InTree = True
    Next
    
    'To extract the critical path you need to start at the Finish node and work backwards via the InLinks
    'Each member of the critical path is marked so it can be drawn in red.
    Set Nd = gNodes(cFINISH)
    Do
        Nd.InLink.Critical = True
        Set Nd = Nd.InLink.Nd1
    Loop Until Nd Is gNodes(cSTART)

End Sub

Public Sub CPClear()
Dim Nodes As Variant, Links As Variant
Dim Index As Integer

    'Reset the flags ready for another run of the Critical Path algorithm.
    If gNodes.Count = 0 Then Exit Sub
    Nodes = gNodes.Items
    For Index = 0 To UBound(Nodes)
        With Nodes(Index)
            Set .InLink = Nothing
            .Distance = 0
        End With
    Next

    'Unmark all the links. These will be Marked up again in the CPMarkup Proc
    Links = gLinks.Items
    For Index = 0 To UBound(Links)
        With Links(Index)
            .InTree = False
            .Critical = False
        End With
    Next

End Sub

Public Sub Main()

    Set mForm = New frmCriticalAnalyser
    Set gLinks = New Dictionary
    Set gNodes = New Dictionary
    
    mForm.Resize = True
    mForm.SetEditMenu
    mForm.Show
    
End Sub
Public Sub DrawCriticalPath()
Dim Nodes As Variant, Links As Variant
Dim X As Single, Y As Single
Dim Index As Integer
Dim Nd As Node
Dim Lk As Link

    'This proc draws the web structure
    
    With mForm
        'Draw the links
        .ForeColor = vbBlack:  .FillColor = .BackColor
        Links = gLinks.Items
        For Index = 0 To UBound(Links)
            Set Lk = Links(Index)
            .ForeColor = vbBlack: .DrawWidth = 1
            If Lk.Critical Then
                .ForeColor = vbRed: .DrawWidth = 2
            ElseIf Lk.InTree Then
                .ForeColor = vbBlue: .DrawWidth = 2
            End If
            mForm.Line (Lk.Nd1.X, Lk.Nd1.Y)-(Lk.Nd2.X, Lk.Nd2.Y)
        Next
        'Draw the nodes
        .DrawWidth = 1
        Nodes = gNodes.Items
        For Index = 0 To UBound(Nodes)
            Set Nd = Nodes(Index)
            .FillColor = vbWhite: .ForeColor = vbBlack
            If Nd Is gNodes(cSTART) Or Nd Is gNodes(cFINISH) Then .FillColor = vbRed: .ForeColor = vbWhite
            mForm.Circle (Nd.X, Nd.Y), cRADIUS
            .PrintText Nd.X - .TextWidth(Nd.Key) / 2, Nd.Y - .TextHeight(Nd.Key) / 2, Nd.Key
        Next
    End With
    
End Sub

Public Sub DrawGantt()
Dim Index As Integer, Cw As Integer, Ch As Integer, MaxKeyLen As Integer, Length As Integer, TotalDist As Integer
Dim Links As Variant, Nodes As Variant
Dim X As Single, Y As Single
Dim Lk As Link
Dim Nd As Node
    
    'This proc draws a grid and overlays red or blue blobs to represent the tasks
    With mForm
        'Work out the positions of the text and the cell height and width (changes as the form resizes)
        .ForeColor = vbBlack
        Ch = (.Height - 500) / (gNodes.Count)
        If Ch < 100 Then Ch = 100
        TotalDist = gNodes(cFINISH).Distance
        Nodes = gNodes.Items
        For Index = 0 To UBound(Nodes)
            Set Nd = Nodes(Index)
            X = 0: Y = 0
            If Not Nd.Key = cSTART And Not Nd.Key = cFINISH Then
                .PrintText 50, Ch * (Index + 1), Nd.Key
                If mForm.TextWidth(Nd.Key) > MaxKeyLen Then MaxKeyLen = mForm.TextWidth(Nd.Key)
            End If
        Next
        MaxKeyLen = MaxKeyLen + 100
        Cw = (.Width - MaxKeyLen) \ (TotalDist + 1)
        If Cw < 100 Then Cw = 100
        .FillColor = &H80000018
        'Blit on the light yellow background.
        BitBlt .hDC, .ScaleX(MaxKeyLen, vbTwips, vbPixels), .ScaleY(Ch, vbTwips, vbPixels), .ScaleX(.Width, vbTwips, vbPixels), .ScaleY(.Height - Ch, vbTwips, vbPixels), .hDC, 0, 0, cBITBLT_PATCOPY
        X = MaxKeyLen
        'Print all the text
        For Index = 1 To TotalDist
            .PrintText X + 25, 25, CStr(Index)
            X = X + Cw
        Next
        'Draw the Tasks in either Red or Blue
        For Index = 0 To UBound(Nodes)
            Set Nd = Nodes(Index)
            If Not Nd.Key = cSTART And Not Nd.Key = cFINISH Then
                X = MaxKeyLen + (Nd.Distance * Cw)
                Y = ((Index + 1) * Ch) + (Ch * 0.25)
                If Not Nd.Links.Count = 0 Then
                    Links = Nd.Links.Items
                    Length = (Links(0).Length * Cw)
                End If
                .FillColor = IIf(Nd.InLink.Critical, vbRed, vbBlue)
                BitBlt .hDC, .ScaleX(X, vbTwips, vbPixels), .ScaleY(Y, vbTwips, vbPixels), .ScaleX(Length, vbTwips, vbPixels), .ScaleY(Ch \ 2, vbTwips, vbPixels), .hDC, 0, 0, cBITBLT_PATCOPY
            End If
        Next
        DrawGrid Ch, Cw, MaxKeyLen
    End With


End Sub


Public Sub FileOpen()
Dim XNodes As IXMLDOMNodeList, XLinks As IXMLDOMNodeList, Elems As IXMLDOMNodeList
Dim XML As DOMDocument, XRoot As IXMLDOMElement
Dim Index As Integer
Dim Nd As Node

    'This proc opens and reads an XML document.
    If gDirtyFile Then
        If MsgBox("Save changes", vbYesNo + vbQuestion) = vbYes Then FileSave
    End If
    
    Clear
    'Create a new XML document object model
    Set XML = New DOMDocument
    XML.async = False
    'Load in the specified XML file
    SetFileName eOpen
    If Len(mFileName) = 0 Then Exit Sub
    XML.Load mFileName
    Set XRoot = XML.documentElement
    'Extract the list of nodes
    Set XNodes = XRoot.selectNodes("Node")
    'Extract the list of links
    Set XLinks = XRoot.selectNodes("Link")
    
    'Set the form's height and width
    With mForm
        .Resize = False
        .Height = XRoot.Attributes(0).Text
        .Width = XRoot.Attributes(1).Text
        .Resize = True
        'Instantiate the Node objects using the XML data
        For Index = 0 To XNodes.Length - 1
            Set Elems = XNodes(Index).childNodes
            Set Nd = New Node
            With Nd
                .Key = Elems(0).Text
                .X = Elems(1).Text
                .Y = Elems(2).Text
                .Length = Elems(3).Text
            End With
            gNodes.Add Nd.Key, Nd
            If IsNumeric(Nd.Key) Then
                If Nd.Key > gNextKey Then gNextKey = Nd.Key
            End If
        Next
        'Instantiate all the links using the XML data
        For Index = 0 To XLinks.Length - 1
            Set Elems = XLinks(Index).childNodes
            .AddLink gNodes(Elems(0).Text), gNodes(Elems(1).Text), gNodes(Elems(0).Text).Length
        Next
        .DirtyWeb = True
        .RefreshDisplay
    End With
    
    gDirtyFile = False
    
End Sub


Public Sub Clear()
    gNodes.RemoveAll
    gLinks.RemoveAll
    gNextKey = 0
End Sub


Public Sub FileSave(Optional Action As eFileAction)
Dim XRoot As IXMLDOMElement, XNd As IXMLDOMElement, XLk As IXMLDOMElement
Dim Nd2 As IXMLDOMElement, Nd1 As IXMLDOMElement, Length As IXMLDOMElement
Dim X As IXMLDOMElement, Y As IXMLDOMElement, Key As IXMLDOMElement
Dim Nodes As Variant, Links As Variant
Dim XML As DOMDocument
Dim Index As Integer
Dim Nd As Node
Dim Lk As Link

    'This proc saves the nodes and links as an XML document.
    'You can open the document using Microsoft's Internet Explorer to see the resulting file structure.
    If Action = eSaveAs Or Len(mFileName) = 0 Then
        SetFileName eSaveAs
        If Len(mFileName) = 0 Then Exit Sub
    End If
    
    'Create a new XML document and add some attributes
    Set XML = New DOMDocument
    XML.async = False
    Set XRoot = XML.createElement("Web")
    XRoot.setAttribute "Height", mForm.Height
    XRoot.setAttribute "Width", mForm.Width
    
    'Create an XML node for every node object
    'Append XML elements to the XML node to hold the object's property values
    Nodes = gNodes.Items
    For Index = 0 To UBound(Nodes)
        Set Nd = Nodes(Index)
        If Not Nd.Key = cSTART And Not Nd.Key = cFINISH Then
            Set XNd = XML.createElement("Node")
            Set Key = XML.createElement("Key")
            Set X = XML.createElement("X")
            Set Y = XML.createElement("Y")
            Set Length = XML.createElement("Length")
            Key.Text = Nd.Key
            X.Text = Nd.X
            Y.Text = Nd.Y
            Length.Text = Nd.Length
            XNd.appendChild Key
            XNd.appendChild X
            XNd.appendChild Y
            XNd.appendChild Length
            XRoot.appendChild XNd
        End If
    Next
    
    'Create an XML node for every link object
    'Append XML elements to the XML node to hold each link's properties
    Links = gLinks.Items
    For Index = 0 To UBound(Links)
        Set Lk = Links(Index)
        If Not Lk.Nd1.Key = cSTART And Not Lk.Nd2.Key = cFINISH Then
            Set XLk = XML.createElement("Link")
            Set Nd1 = XML.createElement("Nd1")
            Set Nd2 = XML.createElement("Nd2")
            Nd1.Text = Lk.Nd1.Key
            Nd2.Text = Lk.Nd2.Key
            XLk.appendChild Nd1
            XLk.appendChild Nd2
            XRoot.appendChild XLk
        End If
    Next
    
    'Append the root element to the document to complete the file structure
    XML.appendChild XRoot
    'Save the XML document
    XML.save mFileName
    gDirtyFile = False
    
End Sub

Public Sub Cleanup()
Dim Nodes, Links, I%
On Error Resume Next
    
    'Get rid of circular references by explicitely cleaning up each node and link object
    Nodes = gNodes.Items()
    For I = 0 To UBound(Nodes)
        Nodes(I).Cleanup
    Next
    Links = Links.Items()
    For I = 0 To UBound(Links)
        Links(I).Cleanup
    Next
    
    'Now it is safe to set the objects to nothing. Everything will now deallocate correctly.
    gNodes.RemoveAll
    gLinks.RemoveAll
    Set gNodes = Nothing
    Set gLinks = Nothing

    Unload mForm
    Set mForm = Nothing
    
End Sub

Public Sub DrawEdit(SelNd1 As Node, SelNd2 As Node)
Dim Nodes As Variant, Links As Variant
Dim X As Single, Y As Single
Dim Index As Integer
Dim Lk As Link
Dim Nd As Node

    With mForm
        .Cls
        DrawGrid cEDITGRID, cEDITGRID
        .ForeColor = vbBlack:  .FillColor = .BackColor: .DrawWidth = 1
        Links = gLinks.Items
        For Index = 0 To UBound(Links)
            Set Lk = Links(Index)
            'Draw the links
            mForm.Line (Lk.Nd1.X, Lk.Nd1.Y)-(Lk.Nd2.X, Lk.Nd2.Y)
            X = (Lk.Nd1.X + Lk.Nd2.X) \ 2
            Y = (Lk.Nd1.Y + Lk.Nd2.Y) \ 2
            'Print the link Lengths
            .DrawStyle = vbInvisible
            mForm.Circle (X, Y), cRADIUS
            .DrawStyle = vbSolid
            .PrintText X - .TextWidth(Lk.Length) / 2, Y - .TextHeight(Lk.Length) / 2, Lk.Length
            'Draw in the link-direction indicator-circles
            X = Lk.Nd1.X + ((Lk.Nd2.X - Lk.Nd1.X)) * 0.75
            Y = Lk.Nd1.Y + ((Lk.Nd2.Y - Lk.Nd1.Y)) * 0.75
            .FillColor = vbGreen
            mForm.Circle (X, Y), cRADIUS \ 5
            .FillColor = .BackColor
        Next
        'Draw the nodes
        .DrawStyle = vbSolid
        Nodes = gNodes.Items
        For Index = 0 To UBound(Nodes)
            Set Nd = Nodes(Index)
            .FillColor = vbWhite: .ForeColor = vbBlack
            If Nd Is SelNd1 Then
                .FillColor = vbGreen
            ElseIf Nd Is SelNd2 Then
                .FillColor = vbRed
            End If
            If Nd Is gNodes(cSTART) Or Nd Is gNodes(cFINISH) Then .FillColor = vbBlue: .ForeColor = vbWhite
            mForm.Circle (Nd.X, Nd.Y), cRADIUS
            .PrintText Nd.X - .TextWidth(Nd.Key) / 2, Nd.Y - .TextHeight(Nd.Key) / 2, Nd.Key
        Next
    End With
    
End Sub



Public Sub DrawGrid(CellH As Integer, CellW As Integer, Optional XOffset As Integer, Optional YOffset As Integer)
Dim Index As Long
Dim OldColor As Long, OldWidth As Long

    'Draw Grid
    With mForm
        OldColor = .ForeColor: OldWidth = .DrawWidth
        .DrawWidth = 1: .ForeColor = &H8000000C
        For Index = XOffset To .Width Step CellW
            mForm.Line (Index, YOffset)-(Index, .Height - 500)
        Next
        For Index = YOffset To .Height - CellH Step CellH
            mForm.Line (0, Index)-(.Width, Index)
        Next
        .DrawWidth = OldWidth: .ForeColor = OldColor
    End With
    
End Sub

Public Sub SnapToGrid(X As Single, Y As Single)
Dim RoundedX As Integer, RoundedY As Integer

    'Rounds the X and Y values so they sit on the grid exactly.
    RoundedX = ((X \ cEDITGRID) * cEDITGRID)
    If X <= RoundedX + cEDITGRID \ 2 Then
        X = RoundedX
    Else
        X = RoundedX + cEDITGRID
    End If
    
    RoundedY = ((Y \ cEDITGRID) * cEDITGRID)
    If Y <= RoundedY + cEDITGRID \ 2 Then
        Y = RoundedY
    Else
        Y = RoundedY + cEDITGRID
    End If
    
End Sub

